#ifndef AUDIO_H
#define AUDIO_H

#include <stdlib.h>

#if RETRO_PLATFORM != RETRO_3DS
#include <vorbis/vorbisfile.h>
#endif

#if RETRO_PLATFORM != RETRO_VITA && RETRO_PLATFORM != RETRO_3DS
#include "SDL.h"
#endif

#define TRACK_COUNT (0x10)
#define SFX_COUNT (0x100)
#define CHANNEL_COUNT (0x4)
#define SFXDATA_COUNT (0x400000)

#define MAX_VOLUME (100)

struct TrackInfo {
    char fileName[0x40];
    bool trackLoop;

#if RETRO_USING_SDLMIXER
    Mix_Music* mus;
#endif
};

struct MusicPlaybackInfo {
    OggVorbis_File vorbisFile;
    int vorbBitstream;
#if RETRO_USING_SDL1_AUDIO
    SDL_AudioSpec spec;
#endif
#if RETRO_USING_SDL2
    SDL_AudioStream *stream;
#endif
    Sint16 *buffer;
    FileInfo fileInfo;
    bool trackLoop;
    uint loopPoint;
    bool loaded;
};

struct SFXInfo {
    char name[0x40];
#if RETRO_USING_SDLMIXER
    Mix_Chunk* chunk;
#endif
    Sint16 *buffer;
    size_t length;
    bool loaded;
};

struct ChannelInfo {
    size_t sampleLength;
    Sint16 *samplePtr;
    int sfxID;
    byte loopSFX;
    sbyte pan;
};

enum MusicStatuses {
    MUSIC_STOPPED = 0,
    MUSIC_PLAYING = 1,
    MUSIC_PAUSED  = 2,
    MUSIC_LOADING = 3,
    MUSIC_READY   = 4,
};

extern int globalSFXCount;
extern int stageSFXCount;

extern int masterVolume;
extern int trackID;
extern int sfxVolume;
extern int bgmVolume;
extern bool audioEnabled;

extern int nextChannelPos;
extern bool musicEnabled;
extern int musicStatus;
extern TrackInfo musicTracks[TRACK_COUNT];
extern SFXInfo sfxList[SFX_COUNT];

extern ChannelInfo sfxChannels[CHANNEL_COUNT];

extern MusicPlaybackInfo musInfo;

#if RETRO_USING_SDLMIXER
extern int currentTrack;
extern int previousTrack;
#endif

#if RETRO_USING_SDL1_AUDIO || RETRO_USING_SDL2
extern SDL_AudioSpec audioDeviceFormat;
#endif
int InitAudioPlayback();
void LoadGlobalSfx();

#if RETRO_USING_SDL1_AUDIO || RETRO_USING_SDL2
void ProcessMusicStream(void *data, Sint16 *stream, int len);
void ProcessAudioPlayback(void *data, Uint8 *stream, int len);
void ProcessAudioMixing(Sint32 *dst, const Sint16 *src, int len, int volume, sbyte pan);


inline void freeMusInfo()
{
    if (musInfo.loaded) {
        SDL_LockAudio();

        if (musInfo.buffer)
            delete[] musInfo.buffer;
#if RETRO_USING_SDL2
        if (musInfo.stream)
            SDL_FreeAudioStream(musInfo.stream);
#endif
        ov_clear(&musInfo.vorbisFile);
        musInfo.buffer       = nullptr;
#if RETRO_USING_SDL2
        musInfo.stream = nullptr;
#endif
        musInfo.trackLoop    = false;
        musInfo.loaded       = false;
        musicStatus          = MUSIC_STOPPED;

        SDL_UnlockAudio();
    }
}
#else
//void ProcessMusicStream() {}
//void ProcessAudioPlayback() {}
//void ProcessAudioMixing() {}

inline void freeMusInfo()
{
    if (musInfo.loaded) {
        SDL_LockAudio();

        if (musInfo.buffer)
            delete[] musInfo.buffer;
        ov_clear(&musInfo.vorbisFile);
        musInfo.buffer    = nullptr;
        musInfo.trackLoop = false;
        musInfo.loopPoint = 0;
        musInfo.loaded    = false;
        musicStatus       = MUSIC_STOPPED;

        SDL_UnlockAudio();
    }
}
#endif

void SetMusicTrack(char *filePath, byte trackID, bool loop);
bool PlayMusic(int track);
inline void StopMusic()
{
#if RETRO_USING_SDLMIXER
    Mix_HaltMusic();
    
    currentTrack = -1;
    previousTrack = -1;
#endif
    printLog("StopMusic() called");

    musicStatus = MUSIC_STOPPED;
    freeMusInfo();
}

void LoadSfx(char *filePath, byte sfxID);
void PlaySfx(int sfx, bool loop);
inline void StopSfx(int sfx)
{
    for (int i = 0; i < CHANNEL_COUNT; ++i) {
        if (sfxChannels[i].sfxID == sfx) {
            MEM_ZERO(sfxChannels[i]);
            sfxChannels[i].sfxID = -1;
        }
    }
}
void SetSfxAttributes(int sfx, int loopCount, sbyte pan);

inline void SetMusicVolume(int volume)
{
#if RETRO_USING_SDLMIXER
    Mix_VolumeMusic(volume);
#endif
     printLog("SetMusicVolume() called");

    if (volume < 0)
        volume = 0;
    if (volume > MAX_VOLUME)
        volume = MAX_VOLUME;
    masterVolume = volume;
}

inline void PauseSound()
{
#if RETRO_USING_SDLMIXER
    Mix_PauseMusic();
#endif
    printLog("PauseSound() called");

    if (musicStatus == MUSIC_PLAYING)
        musicStatus = MUSIC_PAUSED;
}

inline void ResumeSound()
{
#if RETRO_USING_SDLMIXER
    Mix_ResumeMusic();
#endif
    printLog("ResumeSound() called");

    if (musicStatus == MUSIC_PAUSED)
        musicStatus = MUSIC_PLAYING;
}


inline void StopAllSfx()
{
    for (int i = 0; i < CHANNEL_COUNT; ++i) sfxChannels[i].sfxID = -1;
}
inline void ReleaseGlobalSfx()
{
    StopAllSfx();
    for (int i = globalSFXCount - 1; i >= 0; --i) {
        if (sfxList[i].loaded) {
#if RETRO_USING_SDLMIXER
    	    Mix_FreeChunk(sfxList[i].chunk);
#endif

            StrCopy(sfxList[i].name, "");
            free(sfxList[i].buffer);
            sfxList[i].length = 0;
            sfxList[i].loaded = false;
        }
    }
    globalSFXCount = 0;
}
inline void ReleaseStageSfx()
{
    for (int i = stageSFXCount + globalSFXCount; i >= globalSFXCount; --i) {
        if (sfxList[i].loaded) {
#if RETRO_USING_SDLMIXER
	    Mix_FreeChunk(sfxList[i].chunk);
#endif

            StrCopy(sfxList[i].name, "");
            free(sfxList[i].buffer);
            sfxList[i].length = 0;
            sfxList[i].loaded = false;
        }
    }
    stageSFXCount = 0;
}

inline void ReleaseAudioDevice()
{
    StopMusic();
    StopAllSfx();
    ReleaseStageSfx();
    ReleaseGlobalSfx();

#if RETRO_USING_SDLMIXER
    Mix_CloseAudio();
#endif
}

#endif // !AUDIO_H
